/* Copyright (c) 2003, Rincon Networks, Inc.  All rights reserved. */

#include "nlhelp.h"

#include <asm/types.h>
#include <sys/socket.h>
#include <linux/if.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

void rtnl_close(struct rtnl_handle *rth)
{
    close(rth->fd);
}

int rtnl_open(struct rtnl_handle *rth, unsigned subscriptions)
{
    socklen_t addr_len;

    memset(rth, 0, sizeof(rth));

    rth->fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
    if (rth->fd < 0)
        return -1;

    memset(&rth->local, 0, sizeof(rth->local));
    rth->local.nl_family = AF_NETLINK;
    rth->local.nl_groups = subscriptions;

    if (bind(rth->fd, (struct sockaddr*)&rth->local, sizeof(rth->local)) < 0) {
        close(rth->fd);
        return -1;
    }

    addr_len = sizeof(rth->local);
    if (getsockname(rth->fd, (struct sockaddr*)&rth->local, &addr_len) < 0) {
        close(rth->fd);
        return -1;
    }

    if (addr_len != sizeof(rth->local)) {
        close(rth->fd);
        return -1;
    }

    if (rth->local.nl_family != AF_NETLINK) {
        close(rth->fd);
        return -1;
    }

    rth->seq = time(NULL);

    return 0;
}

#ifdef __SH4__

static int __check_dev_name_attr(struct nlmsghdr* hdr,
                                 const char* if_name)
{
    // iterate through message attributes
    void* pv = NLMSG_DATA(hdr);

    if (hdr->nlmsg_len > 0) {
        char rgchDev[IFNAMSIZ + 1];

        int attrlen = hdr->nlmsg_len;
        struct rtattr* attr = (struct rtattr*)pv;

        while (RTA_OK(attr, attrlen)) {
            if (attr->rta_type == RWA_DEV_NAME) {
                // handle the device name attribute
                if (attr->rta_len - RTA_ALIGN(sizeof(struct rtattr)) <=
                    IFNAMSIZ) {
                    memset(rgchDev, 0, sizeof(rgchDev));
                    memcpy(rgchDev, 
                           (char*)attr + RTA_ALIGN(sizeof(struct rtattr)),
                           attr->rta_len - RTA_ALIGN(sizeof(struct rtattr)));
                    if (0 == strcmp(rgchDev, if_name))
                        return 1;
                }
            }

            attr = RTA_NEXT(attr, attrlen);
        }
    }

    return 0;
}

#endif

int rtnl_msg_triggers_dhcp_renew(struct rtnl_handle *rth,
                                 const char* if_name)
{
#ifdef __SH4__
    struct sockaddr_nl sanl;
    socklen_t sanllen = sizeof(sanl);
    
    struct nlmsghdr* hdr;
    int amt;
    char buf[2048];

    amt = recvfrom(rth->fd, buf, sizeof(buf),
                   0, (struct sockaddr*)&sanl, &sanllen);
    if (amt < 0) {
        // some error reading from the netlink socket?
        return 0;
    }

    if (sanl.nl_family != AF_NETLINK) {
        // why is this bound socket returning random address families?
        return 0;
    }

    hdr = (struct nlmsghdr*)buf;
    while (amt >= (int)sizeof(*hdr)) {
        int len = hdr->nlmsg_len;
        int l = len - sizeof(*hdr);

        if(l < 0 || len > amt) {
            // malformed netlink message?
            break;
        }

        // process netlink message
        if (sanl.nl_groups == RTMGRP_Rincon) {
            // handle Rincon-proprietary message
            if (hdr->nlmsg_type == RWM_SSID) {
                // OK, we saw a SSID change message; check to see if it
                // is for 'our' interface; if so, DHCP is retriggered.
                if (__check_dev_name_attr(hdr, if_name))
                    return 1;
            } else if (hdr->nlmsg_type == RWM_MII) {
                // OK, we saw an MII port scan message; trigger DHCP
                return 1;
            }
        } else {
            // where did this come from?
            break;
        }

        len = NLMSG_ALIGN(len);
        amt -= len;
        hdr = (struct nlmsghdr*)((char*)hdr + len);        
    }
#else
    (void)rth;
    (void)if_name;
#endif

    return 0;
}
